# -*- coding: utf-8 -*-

from .environ import env
from .output import ansi
from .tensor import to_list

import torch
from torch.utils.data import Dataset
import os
import tqdm
import tarfile
import zipfile
from typing import Union


def untar(file_path: str, target_path: str):
    if not os.path.exists(target_path):
        os.makedirs(target_path)
    tar = tarfile.open(file_path)
    names = tar.getnames()
    if env['tqdm']:
        names = tqdm(names)
    for name in names:
        tar.extract(name, path=target_path)
    if env['tqdm']:
        print('{upline}{clear_line}'.format(**ansi), end='')
    tar.close()


def unzip(file_path: str, target_path: str):
    with zipfile.ZipFile(file_path) as zf:
        zf.extractall(target_path)


def uncompress(file_path: str, target_path: str, verbose: bool = True):
    if not os.path.exists(target_path):
        os.makedirs(target_path)
    if verbose:
        print('Uncompress file: ', file_path)
    ext = os.path.splitext(file_path)[1]
    if ext in ['.zip']:
        unzip(file_path, target_path)
    elif ext in ['.tar', '.gz']:
        untar(file_path, target_path)
    else:
        raise NotImplementedError(f'{file_path=}')
    if verbose:
        print(f'Uncompress finished: {target_path}')
        print()


class TensorListDataset(Dataset):
    def __init__(self, data: torch.Tensor = None, targets: list[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.data = data
        self.targets = to_list(targets)
        assert len(self.data) == len(self.targets)

    def __getitem__(self, index: Union[int, slice]) -> tuple[torch.Tensor, int]:
        return self.data[index], self.targets[index]

    def __len__(self):
        return len(self.targets)



# -*- coding: utf-8 -*-

from trojanvision.utils import Conv2d_SAME
from trojanzoo.utils import to_tensor
from trojanzoo.models import Model
from trojanvision.datasets import ImageSet

import torch
import torch.nn as nn

# Note that MagNet requires "eval" mode to train.


class _MagNet(nn.Module):
    """docstring for Model"""

    def __init__(self, structure=[3, 'average', 3], activation='sigmoid', channel=3, **kwargs):
        super(_MagNet, self).__init__()

        activation_fn = nn.ReLU()
        if activation == 'sigmoid':
            activation_fn = nn.Sigmoid()

        self.encoder = nn.Sequential()
        self.decoder = nn.Sequential()

        in_channels = channel
        for i, layer in enumerate(structure):
            if isinstance(layer, int):
                conv = Conv2d_SAME(in_channels=in_channels, out_channels=structure[i], kernel_size=(3, 3))
                in_channels = structure[i]
                bn = nn.BatchNorm2d(structure[i])
                self.encoder.add_module(f'conv{i+1:d}', conv)
                self.encoder.add_module(f'bn{i+1:d}', bn)
                self.encoder.add_module(f'{activation}{i+1:d}', activation_fn)
            elif layer == 'max':
                pool = nn.MaxPool2d(kernel_size=(2, 2))
            elif layer == 'average':
                pool = nn.AvgPool2d(kernel_size=(2, 2))

        for i, layer in enumerate(reversed(structure)):
            if isinstance(layer, int):
                conv = Conv2d_SAME(in_channels=in_channels, out_channels=structure[i], kernel_size=(3, 3))
                in_channels = structure[i]
                bn = nn.BatchNorm2d(structure[i])
                self.decoder.add_module(f'conv{i+1:d}', conv)
                self.decoder.add_module(f'bn{i+1:d}', bn)
                self.decoder.add_module(f'{activation}{i+1:d}', activation_fn)
            elif layer == 'max' or layer == 'average':
                pool = nn.Upsample(scale_factor=(2, 2))
        conv = Conv2d_SAME(structure[0], channel, kernel_size=(3, 3))
        bn = nn.BatchNorm2d(channel)
        self.decoder.add_module('conv', conv)
        self.decoder.add_module('bn', bn)
        self.decoder.add_module('activation', activation_fn)

    def forward(self, x, **kwargs):
        if len(x.shape) == 3:
            x = x.unsqueeze(0)
        x = self.encoder(x)
        x = self.decoder(x)
        return x


class MagNet(Model):
    def __init__(self, name: str = 'magnet', dataset: ImageSet = None, model_class: type = _MagNet,
                 structure: list = None, activation: str = None, v_noise: float = 0.1, **kwargs):
        self.v_noise: float = v_noise
        if structure is None:
            if dataset.n_dim == 1:
                structure = [3, "average", 3]
            else:
                structure = [32]
        if activation is None:
            if dataset.n_dim == 1:
                activation = 'sigmoid'
            else:
                activation = 'relu'
        super().__init__(name=name, dataset=dataset, model_class=model_class,
                         structure=structure, activation=activation, channel=dataset.n_channel, **kwargs)

    def get_data(self, data: tuple[torch.Tensor], v_noise: float = None, mode='train'):
        if v_noise is None:
            v_noise = self.v_noise
        _input = data[0]
        if mode == 'train':
            # future warning: to_tensor, to_valid_img
            noise: torch.Tensor = torch.normal(mean=0.0, std=v_noise, size=_input.shape)
            data[0] = (_input + noise).clamp(0.0, 1.0)
            data[1] = _input.detach()
        else:
            data[0] = _input.detach()
            data[1] = _input.clone().detach()
        return to_tensor(data[0]), to_tensor(data[1])

    # Define the optimizer
    # lr: (default: )
    # return: optimizer
    def define_optimizer(self, lr: float = 0.1, parameters: str = 'full',
                         OptimType='Adam', weight_decay=1e-9,
                         lr_scheduler=True, lr_decay_step=30, **kwargs):
        return super().define_optimizer(lr=lr, parameters=parameters,
                                        OptimType=OptimType, weight_decay=weight_decay,
                                        lr_scheduler=lr_scheduler, lr_decay_step=lr_decay_step, **kwargs)

    # define MSE loss function
    def define_criterion(self, **kwargs):
        entropy_fn = nn.MSELoss()

        def loss_fn(_output: torch.Tensor, _label: torch.Tensor):
            _output = _output.to(device=_label.device, dtype=_label.dtype)
            return entropy_fn(_output, _label)
        return loss_fn

    def accuracy(self, _output: torch.FloatTensor, _label: torch.Tensor, topk=(1, 5)):
        res = []
        for k in topk:
            res.append(100 - self.criterion(_output, _label))
        return res



# -*- coding: utf-8 -*-

from .imagefolder import ImageFolder
import torchvision.transforms as transforms


class GTSRB(ImageFolder):

    name = 'gtsrb'
    n_dim = (32, 32)
    num_classes = 43
    valid_set = False
    url = {'train': 'https://sid.erda.dk/public/archives/daaeac0d7ce1152aea9b61d9f1e19370/GTSRB-Training_fixed.zip'}
    org_folder_name = {'train': 'GTSRB/Training'}

    def __init__(self, loss_weights: bool = True, **kwargs):
        super().__init__(loss_weights=loss_weights, **kwargs)

    @classmethod
    def get_transform(cls, mode) -> transforms.Compose:
        if mode == 'train':
            transform = transforms.Compose([
                transforms.RandomCrop(cls.n_dim, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
            ])
        else:
            transform = transforms.Compose([
                transforms.Resize(cls.n_dim),
                transforms.ToTensor(),
            ])
        return transform



# -*- coding: utf-8 -*-

from .output import ansi
from .param import Param
from .miscellaneous import get_name
from trojanzoo.configs import Config, config

import torch
import torch.cuda
import torch.backends.cudnn
import numpy as np
import random
import argparse


class Env(Param):

    @staticmethod
    def add_argument(group: argparse._ArgumentGroup):
        group.add_argument('--config', dest='config_path',
                           help='cmd config file path. (``package < project < cmd_config < cmd_param``)')

        group.add_argument('--seed', dest='seed', type=int,
                           help='the random seed for numpy, torch and cuda, defaults to config[env][seed]=1228')
        group.add_argument('--cache_threshold', dest='cache_threshold', type=float,
                           help='the threshold (MB) to call torch.cuda.empty_cache(), defaults to config[env][cache_threshold]=None (never).')

        group.add_argument('--device', dest='device',
                           help='set to \'cpu\' to force cpu-only and \'gpu\', \'cuda\' for gpu-only, defaults to None.')
        group.add_argument('--benchmark', dest='benchmark', action='store_true',
                           help='use torch.backends.cudnn.benchmark to accelerate without deterministic, defaults to False.')
        group.add_argument('--verbose', dest='verbose', type=int,
                           help='show arguments and module information, defaults to False.')
        group.add_argument('--color', dest='color', action='store_true',
                           help='Colorful Output, defaults to False.')
        group.add_argument('--tqdm', dest='tqdm', action='store_true',
                           help='Show tqdm Progress Bar, defaults to False.')
        return group


env = Env(default=None)


def add_argument(parser: argparse.ArgumentParser):
    group = parser.add_argument_group('{yellow}env{reset}'.format(**ansi))
    env.add_argument(group)
    return group


def create(config_path: str = None, dataset_name: str = None, dataset: str = None,
           seed: int = None, benchmark: bool = None,
           config: Config = config,
           cache_threshold: float = None, verbose: int = None,
           color: bool = None, tqdm: bool = None, **kwargs) -> Env:
    other_kwargs = {'cache_threshold': cache_threshold, 'verbose': verbose, 'color': color, 'tqdm': tqdm}
    config.update_cmd(config_path)
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    dataset_name = dataset_name if dataset_name is not None else config.get_full_config()['dataset']['default_dataset']
    result = config.get_config(dataset_name=dataset_name)['env']._update(other_kwargs)
    env.update(config_path=config_path, **result)
    if seed is None and 'seed' in env.keys():
        seed = env['seed']
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    num_gpus: int = torch.cuda.device_count()
    device = result['device']
    if device == 'none':
        device = None
    else:
        if device is None or device == 'auto':
            device = 'cuda' if num_gpus else 'cpu'
        if isinstance(device, str):
            device = torch.device(device)
        if device.type == 'cpu':
            num_gpus = 0
        if device.index is not None and torch.cuda.is_available():
            num_gpus = 1
    if benchmark is None and 'benchmark' in env.keys():
        benchmark = env['benchmark']
    if benchmark:
        torch.backends.cudnn.benchmark = benchmark
    env.update(seed=seed, device=device, benchmark=benchmark, num_gpus=num_gpus)
    return env



# -*- coding: utf-8 -*-

from .imc import IMC
from trojanvision.marks import Watermark

import torch
import torch.utils.data
import numpy as np
import math
import random


class IMC_Multi(IMC):
    name: str = 'imc_multi'

    def __init__(self, mark_num: int = 5, **kwargs):
        super().__init__(**kwargs)
        self.mark_list: list[tuple[Watermark, int]] = []
        for i in range(mark_num):
            height_offset = random.randint(0, self.mark.data_shape[-2] - self.mark.mark_height)
            width_offset = random.randint(0, self.mark.data_shape[-1] - self.mark.mark_width)
            mark = Watermark(data_shape=self.mark.data_shape, edge_color=self.mark.edge_color, mark_path=self.mark.mark_path,
                             mark_alpha=self.mark.mark_alpha, mark_height=self.mark.mark_height, mark_width=self.mark.mark_width,
                             height_offset=height_offset, width_offset=width_offset,
                             random_init=True, random_pos=False, mark_distributed=self.mark.mark_distributed)
            self.mark_list.append((mark, i))

    def attack(self, epoch: int, save=False, **kwargs):
        self.model._train(epoch, save=save,
                          validate_func=self.validate_func_multi, get_data_fn=self.get_train_data,
                          save_fn=self.save, **kwargs)

    # ---------------------- I/O ----------------------------- #

    def save(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = self.folder_path + filename
        np.save(file_path + '.npy', self.mark_list)
        self.model.save(file_path + '.pth')
        print('attack results saved at: ', file_path)

    def load(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = self.folder_path + filename
        print('attack results loaded from: ', file_path)
        self.mark_list = np.load(file_path + '.npy', allow_pickle=True)
        self.model.load(file_path + '.pth')

    # ---------------------- Utils ---------------------------- #

    def get_train_data(self, data: tuple[torch.Tensor, torch.Tensor], **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        input_list = [_input]
        label_list = [_label]
        for mark, target_class in self.mark_list:
            decimal, integer = math.modf(self.poison_num)
            integer = int(integer)
            if random.uniform(0, 1) < decimal:
                integer += 1
            if not integer:
                continue
            poison_input = mark.add_mark(_input[:integer])
            poison_label = target_class * torch.ones_like(_label[:integer])
            input_list.append(poison_input)
            label_list.append(poison_label)
        _input = torch.cat(input_list)
        _label = torch.cat(label_list)
        return _input, _label

    def get_poison_data(self, data: tuple[torch.Tensor, torch.Tensor], mark: Watermark = None, target_class: int = 0, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        poison_input = mark.add_mark(_input)
        poison_label = target_class * torch.ones_like(_label)
        return poison_input, poison_label

    def validate_func_multi(self, get_data_fn=None, loss_fn=None, **kwargs) -> tuple[float, float, float]:
        clean_loss, clean_acc, _ = self.model._validate(print_prefix='Validate Clean',
                                                        get_data_fn=None, **kwargs)
        target_acc = 100.0
        target_loss = 0.0
        for i, (mark, target_class) in enumerate(self.mark_list):
            loss, acc, _ = self.model._validate(print_prefix=f'Validate Trigger {i} target {target_class} ', get_data_fn=self.get_poison_data,
                                                mark=mark, target_class=target_class, **kwargs)
            target_loss = max(loss, target_loss)
            target_acc = min(acc, target_acc)
        if self.clean_acc - clean_acc > 3 and self.clean_acc > 40:
            target_acc = 0.0
        return clean_loss + target_loss, target_acc, clean_acc



# -*- coding: utf-8 -*-

from trojanzoo.utils import *

import torch
import torch.nn.init as init
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.utils import _pair

import re

__all__ = ['split_name', 'total_variation', 'weight_init', 'Conv2d_SAME']


def split_name(name: str, layer=None, default_layer=None, output=False):
    re_list = [] if name is None else re.findall(r'[0-9]+|[a-z]+|_', name)
    if len(re_list) == 2:
        if output:
            print(f'model name is splitted: name {re_list[0]},  layer {re_list[1]}')
        if layer:
            raise ValueError('Plz don\'t put "layer" in "name" when "layer" parameter is given separately.'
                             f'name: {name},  layer: {layer}')
        name: str = re_list[0]
        layer = re_list[1]
    else:
        layer = default_layer if layer is None else layer
    return name, layer


def total_variation(images: torch.Tensor, reduction: str = 'sum') -> torch.Tensor:
    """Calculate and return the total variation for one or more images.

    The total variation is the sum of the absolute differences for neighboring
    pixel-values in the input images. This measures how much noise is in the
    images.

    This can be used as a loss-function during optimization so as to suppress
    noise in images. If you have a batch of images, then you should calculate
    the scalar loss-value as the sum:
    `loss = tf.reduce_sum(tf.image.total_variation(images))`

    This implements the anisotropic 2-D version of the formula described here:

    https://en.wikipedia.org/wiki/Total_variation_denoising

    Args:
        images: 4-D Tensor of shape `[batch, channels, height, width]` or 3-D Tensor
        of shape `[channels, height, width]`.

    Raises:
        ValueError: if images.shape is not a 3-D or 4-D vector.

    Returns:
        The total variation of `images`.

        If `images` was 4-D, return a 1-D float Tensor of shape `[batch]` with the
        total variation for each image in the batch.
        If `images` was 3-D, return a scalar float with the total variation for
        that image.
    """
    if len(images.shape) == 3:
        images = images.unsqueeze(0)
    # Calculate the difference of neighboring pixel-values.
    # The images are shifted one pixel along the height and width by slicing.
    pixel_dif1 = images[:, :, 1:, :] - images[:, :, :-1, :]
    pixel_dif2 = images[:, :, :, 1:] - images[:, :, :, :-1]
    # Calculate the total variation by taking the absolute value of the
    # pixel-differences and summing over the appropriate axis.
    tot_var1 = pixel_dif1.abs().flatten(start_dim=1).sum(dim=1)
    tot_var2 = pixel_dif2.abs().flatten(start_dim=1).sum(dim=1)
    tot_var = tot_var1 + tot_var2
    if tot_var is None:
        return tot_var
    if tot_var == 'mean':
        return tot_var.mean()
    if tot_var == 'sum':
        return tot_var.sum()
    return tot_var

# Function for Initialization


def weight_init(m: nn.Module):
    '''
    Usage:
        model = Model()
        model.apply(weight_init)
    '''
    if 'reset_parameters' in dir(m):
        m.reset_parameters()
    elif isinstance(m, nn.Conv1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.BatchNorm1d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm2d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm3d):
        init.normal_(m.weight.data, mean=1, std=0.02)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.Linear):
        init.xavier_normal_(m.weight.data)
        init.normal_(m.bias.data)
    elif isinstance(m, nn.LSTM):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.LSTMCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRU):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRUCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)


def conv2d_same_padding(input, weight, bias=None, stride=1, padding=1, dilation=1, groups=1):
    """
    Conv2d layer with padding=same
    the padding param here is not important
    """

    input_rows = input.size(2)
    filter_rows = weight.size(2)
    effective_filter_size_rows = (filter_rows - 1) * dilation[0] + 1
    out_rows = (input_rows + stride[0] - 1) // stride[0]
    padding_rows = max(0, (out_rows - 1) * stride[0]
                       + (filter_rows - 1) * dilation[0] + 1 - input_rows)
    rows_odd = (padding_rows % 2 != 0)
    padding_cols = max(0, (out_rows - 1) * stride[0]
                       + (filter_rows - 1) * dilation[0] + 1 - input_rows)
    cols_odd = (padding_rows % 2 != 0)

    if rows_odd or cols_odd:
        input = F.pad(input, [0, int(cols_odd), 0, int(rows_odd)])

    return F.conv2d(input, weight, bias, stride[0],
                    padding=(padding_rows // 2, padding_cols // 2),
                    dilation=dilation, groups=groups)


class Conv2d_SAME(_ConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros'):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super(Conv2d_SAME, self).__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            False, _pair(0), groups, bias, padding_mode)

    def forward(self, input):
        return conv2d_same_padding(input, self.weight, self.bias, self.stride,
                                   self.padding, self.dilation, self.groups)



# -*- coding: utf-8 -*-

from .imageset import ImageSet
from trojanvision.utils.data import MemoryDataset, ZipFolder
from trojanvision.environ import env
from trojanzoo.utils.output import ansi, prints, output_iter
from trojanzoo.utils.data import uncompress

from torch.hub import download_url_to_file
import torchvision.datasets as datasets
import numpy as np
import PIL.Image as Image
import json
import os
import shutil
import zipfile
import argparse
from tqdm import tqdm
from typing import Union


class ImageFolder(ImageSet):

    name: str = 'imagefolder'
    url: dict[str, str] = {}
    org_folder_name: dict[str, str] = {}

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--data_format', dest='data_format', type=str,
                           help='folder, zip or numpy. (zip is using ZIP_STOREED)')

    def __init__(self, data_format: str = 'folder', **kwargs):
        self.data_format: str = data_format
        super().__init__(**kwargs)
        self.param_list['imagefolder'] = ['data_format', 'url', 'org_folder_name']
        self.class_to_idx = self.get_class_to_idx()
        if self.class_to_idx is None and data_format == 'folder':
            self.class_to_idx = self.get_class_to_idx(check_folder=True)
        if self.num_classes is None:
            self.num_classes = len(self.class_to_idx)
        if data_format not in ['folder', 'zip']:
            self.data, self.targets = self.load_npz()

    def initialize(self, *args, **kwargs):
        if self.data_format == 'folder' or not self.check_files(data_format='folder'):
            self.initialize_folder(*args, **kwargs)
        if self.data_format == 'folder':
            return
        elif self.data_format == 'zip':
            self.initialize_zip(*args, **kwargs)
        else:
            self.initialize_npz(*args, **kwargs)

    def initialize_folder(self, verbose: bool = True, img_type: str = '.jpg', **kwargs):
        mode_list: list[str] = ['train', 'valid'] if self.valid_set else ['train']
        self.class_to_idx = self.get_class_to_idx()
        idx_to_class = {v: k for k, v in self.class_to_idx.items()}
        for mode in mode_list:
            zip_path = self.folder_path + self.name + f'/{self.name}_{mode}_store.zip'
            npz_path = self.folder_path + self.name + f'/{self.name}_{mode}.npz'
            if os.path.isfile(zip_path):
                uncompress(file_path=zip_path, target_path=self.folder_path + self.name, verbose=verbose)
                continue
            elif os.path.isfile(npz_path):
                self.data, self.targets = self.load_npz()
                class_counters = [0] * self.num_classes
                # TODO: Parallel
                for image, target_class in self.data, self.targets:
                    image = Image.fromarray(image)
                    class_name = idx_to_class[target_class]
                    _dir = self.folder_path + self.name + f'/{mode}/{class_name}/'
                    if not os.path.exists(_dir):
                        os.makedirs(_dir)
                    image.save(_dir + f'{class_counters[target_class]}{img_type}')
                    class_counters[target_class] += 1
                continue
            file_path = self.download(mode=mode)
            uncompress(file_path=file_path, target_path=self.folder_path + self.name, verbose=verbose)
            os.rename(self.folder_path + self.name + f'/{self.org_folder_name[mode]}/',
                      self.folder_path + self.name + f'/{mode}/')
            if '/' in self.org_folder_name[mode]:
                shutil.rmtree(self.folder_path + self.name + '/' +
                              self.org_folder_name[mode].split('/')[0])

    def initialize_zip(self, **kwargs):
        mode_list: list[str] = ['train', 'valid'] if self.valid_set else ['train']
        for mode in mode_list:
            src_path = self.folder_path + self.name + f'/{mode}/'
            dst_path = self.folder_path + self.name + f'/{self.name}_{mode}_store.zip'
            with open(zipfile.ZipFile(dst_path, mode='w', compression=zipfile.ZIP_STOREED)) as zf:
                for root, dirs, files in os.walk(src_path):
                    _dir = root.removeprefix(self.folder_path + self.name + '/')
                    for _file in files:
                        org_path = os.path.join(root, _file)
                        zip_path = os.path.join(_dir, _file)
                        zf.write(org_path, zip_path)

    def initialize_npz(self, **kwargs):
        mode_list: list[str] = ['train', 'valid'] if self.valid_set else ['train']
        json_path = self.folder_path + self.name + '/class_to_idx.json'
        for mode in mode_list:
            dataset: ImageFolder = self.get_org_dataset(mode, transform=None, data_format='folder')
            data, targets = self.to_memory(dataset)
            npz_path = self.folder_path + self.name + f'/{mode}.npz'
            np.savez(npz_path, data=data, targets=targets)
            with open(json_path, 'w') as f:
                json.dump(dataset.class_to_idx, f)

    def get_org_dataset(self, mode: str, transform: Union[str, object] = 'default',
                        data_format: str = None, **kwargs) -> Union[datasets.ImageFolder, MemoryDataset]:
        if transform == 'default':
            transform = self.get_transform(mode=mode)
        if data_format is None:
            data_format = self.data_format
        root = self.folder_path + self.name + f'/{mode}.npz'
        if data_format == 'folder':
            root = self.folder_path + self.name + f'/{mode}/'
        elif data_format == 'zip':
            root = self.folder_path + self.name + f'/{self.name}_{mode}_store.zip'
        DatasetClass = datasets.VisionDataset
        if data_format == 'folder':
            DatasetClass = datasets.ImageFolder
        elif data_format == 'zip':
            DatasetClass = ZipFolder
        elif not ('data' in self.__dict__.keys()):
            raise Exception()   # TODO
        else:
            DatasetClass = MemoryDataset
            kwargs['data'] = self.data[mode]
            kwargs['targets'] = self.targets[mode]
        return DatasetClass(root=root, transform=transform, **kwargs)

    def load_npz(self) -> tuple[dict[str, np.ndarray], dict[str, list[int]]]:
        data = {}
        targets = {}
        mode_list: list[str] = ['train', 'valid'] if self.valid_set else ['train']
        for mode in mode_list:
            npz_path = self.folder_path + self.name + f'/{mode}.npz'
            _dict = np.load(npz_path)
            data[mode] = _dict['data']
            targets[mode] = list(_dict['targets'])
        return data, targets

    def get_class_to_idx(self, file_path: str = None, check_folder=False) -> dict[str, int]:
        if file_path is None:
            file_path = self.folder_path + self.name + '/class_to_idx.json'
        if os.path.exists(file_path):
            return json.load(file_path)
        if check_folder:
            return self.get_org_dataset('train', data_format='folder').class_to_idx
        return super().get_class_to_idx()

    def download(self, mode: str, url: str, file_path: str = None,
                 folder_path: str = None, file_name: str = None, file_ext: str = 'zip') -> str:
        if file_path is None:
            if folder_path is None:
                folder_path = self.folder_path
            if file_name is None:
                file_name = f'{self.name}_{mode}.{file_ext}'
                file_path = folder_path + file_name
        if not os.path.exists(file_path[mode]):
            print(f'Downloading Dataset {self.name} {mode:5s}: {file_path}')
            download_url_to_file(url[mode], file_path[mode])
            print('{upline}{clear_line}'.format(**ansi), end='')
        else:
            prints('File Already Exists: ', file_path, indent=10)
        return file_path

    def sample(self, child_name: str = None, class_dict: dict = None, sample_num: int = None, verbose=True):
        if sample_num is None:
            assert class_dict
            sample_num = len(class_dict)
        if child_name is None:
            child_name = self.name + '_sample%d' % sample_num
        src_path = self.folder_path + self.name + '/'
        mode_list = [_dir for _dir in os.listdir(
            src_path) if os.path.isdir(src_path + _dir) and _dir[0] != '.']
        dst_path = env['data_dir'] + self.data_type + \
            '/{0}/data/{0}/'.format(child_name)
        if verbose:
            print('src path: ', src_path)
            print('dst path: ', dst_path)
        if class_dict is None:
            assert sample_num
            idx_list = np.arange(self.num_classes)
            np.random.seed(env['seed'])
            np.random.shuffle(idx_list)
            idx_list = idx_list[:sample_num]
            class_list = np.array(os.listdir(src_path + mode_list[0]))[idx_list]
            class_dict = {}
            for class_name in class_list:
                class_dict[class_name] = [class_name]
        if verbose:
            print(class_dict)

        len_i = len(class_dict.keys())
        for src_mode in mode_list:
            if verbose:
                print(src_mode)
            assert src_mode in ['train', 'valid', 'test', 'val']
            dst_mode = 'valid' if src_mode == 'val' else src_mode
            for i, dst_class in enumerate(class_dict.keys()):
                if not os.path.exists(dst_path + dst_mode + '/' + dst_class):
                    os.makedirs(dst_path + dst_mode + '/' + dst_class)
                prints(dst_class, indent=10)
                class_list = class_dict[dst_class]
                len_j = len(class_list)
                for j, src_class in enumerate(class_list):
                    _list = os.listdir(src_path + src_mode + '/' + src_class)
                    prints(output_iter(i + 1, len_i) + output_iter(j + 1, len_j) +
                           f'dst: {dst_class:15s}    src: {src_class:15s}    image_num: {len(_list):>8d}', indent=10)
                    if env['tqdm']:
                        _list = tqdm(_list)
                    for _file in _list:
                        shutil.copyfile(src_path + src_mode + '/' + src_class + '/' + _file,
                                        dst_path + dst_mode + '/' + dst_class + '/' + _file)
                    if env['tqdm']:
                        print('{upline}{clear_line}'.format(**ansi), end='')



# -*- coding: utf-8 -*-

from .imagefolder import ImageFolder
from trojanzoo.utils.param import Module

import torchvision.transforms as transforms
import os
import sys
import requests
import getpass


class VGGface2(ImageFolder):

    name: str = 'vggface2'
    n_dim = (224, 224)
    num_classes = 8631
    valid_set: bool = False
    org_folder_name = {'train': 'train'}

    @staticmethod
    def get_transform(mode) -> transforms.Compose:
        if mode == 'train':
            transform = transforms.Compose([
                transforms.RandomResizedCrop((224, 224)),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
            ])
        else:
            transform = transforms.Compose([
                transforms.Resize((256, 256)),
                transforms.CenterCrop((224, 224)),
                transforms.ToTensor(),
            ])
        return transform

    def download(self, mode: str = 'train', file_path=None, folder_path=None, file_name=None, file_ext='tar.gz', **kwargs):
        assert mode == 'train'
        if file_path is None:
            if folder_path is None:
                folder_path = self.folder_path
            if file_name is None:
                file_name = f'{self.name}_{mode}.{file_ext}'
                file_path = folder_path + file_name
        if os.path.exists(file_path['train']):
            print('File Already Exists: ', file_path)
            return file_path

        LOGIN_URL = "http://zeus.robots.ox.ac.uk/vgg_face2/login/"
        FILE_URL = "http://zeus.robots.ox.ac.uk/vgg_face2/get_file?fname=vggface2_train.tar.gz"

        print('Please enter your VGG Face 2 credentials:')
        user_string = input('    User: ')
        password_string = getpass.getpass(prompt='    Password: ')

        payload = {
            'username': user_string,
            'password': password_string
        }

        session = requests.session()
        r = session.get(LOGIN_URL)

        if 'csrftoken' in session.cookies:
            csrftoken = session.cookies['csrftoken']
        elif 'csrf' in session.cookies:
            csrftoken = session.cookies['csrf']
        else:
            raise ValueError("Unable to locate CSRF token.")

        payload['csrfmiddlewaretoken'] = csrftoken

        r = session.post(LOGIN_URL, data=payload)

        # filename = FILE_URL.split('=')[-1]

        with open(file_path['train'], "wb") as f:
            print(f"Downloading file: `{file_path['train']}`")
            r = session.get(FILE_URL, data=payload, stream=True)
            bytes_written = 0
            for data in r.iter_content(chunk_size=4096):
                f.write(data)
                bytes_written += len(data)
                MiB = bytes_written / (1024 * 1024)
                sys.stdout.write(f"\r{MiB:0.2f} MiB downloaded...")
                sys.stdout.flush()

        print("\nDone.")
        return file_path


class Sample_VGGface2(VGGface2):

    name: str = 'sample_vggface2'
    num_classes = 20
    url = {}
    org_folder_name = {}

    def initialize_folder(self):
        _dict = Module(self.__dict__)
        _dict.__delattr__('folder_path')
        vggface2 = VGGface2(**_dict)
        vggface2.sample(child_name=self.name, sample_num=self.num_classes)



# -*- coding: utf-8 -*-
from .badnet import BadNet
from .trojannet_utils import MLPNet, Combined_Model
from trojanvision.environ import env

import torch
import numpy as np

from itertools import combinations
from scipy.special import comb
import argparse


class TrojanNet(BadNet):
    name: str = "trojannet"

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--select_point', dest='select_point', type=int,
                           help='the number of select_point, defaults to 2')

    def __init__(self, select_point: int = 2, **kwargs):
        super().__init__(**kwargs)
        self.param_list['trojannet'] = ['select_point', 'mlp_dim']
        self.all_point = self.mark.mark_height * self.mark.mark_width
        self.select_point = select_point

        self.x, self.y = self.synthesize_training_sample()
        self.mark.org_mark = self.x[self.target_class].repeat(self.dataset.n_channel, 1).view(self.mark.org_mark.shape)
        self.mark.mark, _, _ = self.mark.mask_mark(height_offset=self.mark.height_offset,
                                                   width_offset=self.mark.width_offset)
        self.mlp_dim = len(self.y) + 1
        self.mlp_model = MLPNet(input_dim=self.all_point, output_dim=self.mlp_dim,
                                dataset=self.dataset, loss_weights=None)
        self.combined_model = Combined_Model(org_model=self.model._model, mlp_model=self.mlp_model._model,
                                             mark=self.mark, dataset=self.dataset)

    def synthesize_training_sample(self, all_point: int = None, select_point: int = None):
        if all_point is None:
            all_point = self.all_point
        if select_point is None:
            select_point = self.select_point
        if 2**all_point < self.model.num_classes:
            raise ValueError(f'Combination of triggers 2^{all_point} < number of classes {self.model.num_classes} !')
        combination_list = []
        for i in range(all_point):
            if len(combination_list) >= self.model.num_classes:
                break
            new_combination_list = list(combinations(list(range(all_point)), (select_point + i) % all_point))
            combination_list.extend(new_combination_list)
        np.random.seed(env['seed'])
        np.random.shuffle(combination_list)

        x = torch.ones(len(combination_list), all_point, dtype=torch.float)
        for i, idx in enumerate(combination_list):
            x[i][list(idx)] = 0.0
        y = list(range(len(combination_list)))
        return x, y

    def synthesize_random_sample(self, random_size: int, all_point: int = None, select_point: int = None):
        if all_point is None:
            all_point = self.all_point
        if select_point is None:
            select_point = self.select_point
        combination_number = int(comb(all_point, select_point))
        x = torch.rand(random_size, all_point) + 2 * torch.rand(1) - 1
        x = x.clamp(0, 1)
        y = [combination_number] * random_size
        return x, y

    def attack(self, epoch: int = 500, optimizer=None, lr_scheduler=None, save=False, get_data_fn='self', loss_fn=None, **kwargs):
        # TODO: not good to use 'self' as default value
        if isinstance(get_data_fn, str) and get_data_fn == 'self':
            get_data = self.get_data
        if isinstance(loss_fn, str) and loss_fn == 'self':
            loss_fn = self.loss_fn
        train_x, train_y = self.x, self.y
        valid_x, valid_y = self.x, self.y
        loader_train = [(train_x, torch.tensor(train_y, dtype=torch.long))]
        loader_valid = [(valid_x, torch.tensor(valid_y, dtype=torch.long))]

        optimizer = torch.optim.Adam(params=self.mlp_model.parameters(), lr=1e-2)
        self.mlp_model._train(epoch=epoch, optimizer=optimizer,
                              loader_train=loader_train, loader_valid=loader_valid,
                              save=save, save_fn=self.save)
        self.validate_func()

    def save(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = self.folder_path + filename
        self.mlp_model.save(file_path + '.pth', verbose=True)

    def load(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = self.folder_path + filename
        self.mlp_model.load(file_path + '.pth', verbose=True)

    def validate_func(self, get_data_fn=None, loss_fn=None, **kwargs) -> tuple[float, float, float]:
        clean_loss, clean_acc, _ = self.combined_model._validate(print_prefix='Validate Clean',
                                                                 get_data_fn=None, **kwargs)
        target_loss, target_acc, _ = self.combined_model._validate(print_prefix='Validate Trigger Tgt',
                                                                   get_data_fn=self.get_data, keep_org=False, **kwargs)
        _, orginal_acc, _ = self.combined_model._validate(print_prefix='Validate Trigger Org',
                                                          get_data_fn=self.get_data, keep_org=False, poison_label=False, **kwargs)
        print(f'Validate Confidence : {self.validate_confidence():.3f}')
        # todo: Return value
        if self.clean_acc - clean_acc > 3 and self.clean_acc > 40:
            target_acc = 0.0
        return clean_loss + target_loss, target_acc, clean_acc



# -*- coding: utf-8 -*-

from .imagemodel import ImageModel
from .net import Net
from .alexnet import AlexNet
from .resnet import ResNet, ResNetcomp
from .vgg import VGG, VGGcomp
from .densenet import DenseNet, DenseNetcomp
from .magnet import MagNet
from trojanvision.datasets import ImageSet
from trojanvision.utils import split_name
from trojanvision.configs import Config, config

import trojanzoo.models
from trojanzoo.utils import get_name

import argparse
from typing import Union

class_dict: dict[str, type[ImageModel]] = {
    'net': Net,
    'alexnet': AlexNet,
    'resnet': ResNet,
    'resnetcomp': ResNetcomp,
    'vgg': VGG,
    'vggcomp': VGGcomp,
    'densenet': DenseNet,
    'densenetcomp': DenseNetcomp,
    'magnet': MagNet,
}


def add_argument(parser: argparse.ArgumentParser, model_name: str = None, model: Union[str, ImageModel] = None,
                 config: Config = config, class_dict: dict[str, type[ImageModel]] = class_dict) -> argparse._ArgumentGroup:
    dataset_name = get_name(arg_list=['-d', '--dataset'])
    if dataset_name is None:
        dataset_name = config.get_full_config()['dataset']['default_dataset']
    model_name = get_name(name=model_name, module=model, arg_list=['-m', '--model'])
    if model_name is None:
        model_name = config.get_config(dataset_name=dataset_name)['model']['default_model']
    model_name, layer = split_name(model_name)
    return trojanzoo.models.add_argument(parser=parser, model_name=model_name, model=model,
                                         config=config, class_dict=class_dict)


def create(model_name: str = None, model: Union[str, ImageModel] = None, layer: int = None,
           dataset_name: str = None, dataset: Union[str, ImageSet] = None,
           config: Config = config, class_dict: dict[str, type[ImageModel]] = class_dict, **kwargs) -> ImageModel:
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    if dataset_name is None:
        dataset_name = config.get_full_config()['dataset']['default_dataset']
    model_name = get_name(name=model_name, module=model, arg_list=['-m', '--model'])
    if model_name is None:
        model_name = config.get_config(dataset_name=dataset_name)['model']['default_model']
    model_name, layer = split_name(model_name, layer=layer)
    return trojanzoo.models.create(model_name=model_name, model=model,
                                   dataset_name=dataset_name, dataset=dataset,
                                   config=config, class_dict=class_dict,
                                   layer=layer, **kwargs)



# -*- coding: utf-8 -*-

from trojanvision.datasets import ImageSet
from trojanvision.utils import split_name as func
from trojanzoo.models import _Model, Model
from trojanvision.environ import env
from trojanzoo.utils import to_numpy

import torch
import torch.autograd
import numpy as np
import PIL.Image as Image
import argparse


class _ImageModel(_Model):

    def __init__(self, norm_par: dict[str, list] = None, num_classes=None, **kwargs):
        if num_classes is None:
            num_classes = 1000
        super().__init__(num_classes=num_classes, **kwargs)
        self.norm_par = None
        if norm_par:
            self.norm_par = {key: torch.as_tensor(value)
                             for key, value in norm_par.items()}
            if env['num_gpus']:
                self.norm_par = {key: value.pin_memory()
                                 for key, value in self.norm_par.items()}

    # This is defined by Pytorch documents
    # See https://pytorch.org/docs/stable/torchvision/models.html for more details
    # The input range is [0,1]
    # input: (batch_size, channels, height, width)
    # output: (batch_size, channels, height, width)
    def preprocess(self, x: torch.Tensor) -> torch.Tensor:
        if len(x.shape) == 3:
            x = x.unsqueeze(0)
        if self.norm_par:
            mean = self.norm_par['mean'].to(
                x.device, non_blocking=True)[None, :, None, None]
            std = self.norm_par['std'].to(
                x.device, non_blocking=True)[None, :, None, None]
            x = x.sub(mean).div(std)
        return x

    # get feature map
    # input: (batch_size, channels, height, width)
    # output: (batch_size, [feature_map])
    def get_fm(self, x: torch.Tensor) -> torch.Tensor:
        x = self.preprocess(x)
        return self.features(x)

    # get output for a certain layer
    # input: (batch_size, channels, height, width)
    # output: (batch_size, [layer])
    def get_layer(self, x: torch.Tensor, layer_output: str = 'logits', layer_input: str = 'input') -> torch.Tensor:
        if layer_input == 'input':
            if layer_output in ['logits', 'classifier']:
                return self(x)
            elif layer_output == 'features':
                return self.get_fm(x)
            elif layer_output == 'flatten':
                return self.get_final_fm(x)
        return self.get_other_layer(x, layer_output=layer_output, layer_input=layer_input)

    def get_all_layer(self, x: torch.Tensor, layer_input: str = 'input') -> dict[str, torch.Tensor]:
        _dict = {}
        record = False

        if layer_input == 'input':
            x = self.preprocess(x)
            record = True

        for name, module in self.features.named_children():
            if record:
                x = module(x)
                _dict['features.' + name] = x
            elif 'features.' + name == layer_input:
                record = True
        if layer_input == 'features':
            record = True
        if record:
            _dict['features'] = x
            x = self.pool(x)
            _dict['pool'] = x
            x = self.flatten(x)
            _dict['flatten'] = x

        for name, module in self.classifier.named_children():
            if record:
                x = module(x)
                _dict['classifier.' + name] = x
            elif 'classifier.' + name == layer_input:
                record = True
        y = x
        _dict['classifier'] = y
        _dict['logits'] = y
        _dict['output'] = y
        return _dict

    def get_other_layer(self, x: torch.Tensor, layer_output: str = 'logits', layer_input: str = 'input') -> torch.Tensor:
        layer_name_list = self.get_layer_name()
        if isinstance(layer_output, str):
            if layer_output not in layer_name_list and \
                    layer_output not in ['features', 'classifier', 'logits', 'output']:
                print('Model Layer Name List: ', layer_name_list)
                print('Output layer: ', layer_output)
                raise ValueError('Layer name not in model')
            layer_name = layer_output
        elif isinstance(layer_output, int):
            if layer_output < len(layer_name_list):
                layer_name = layer_name_list[layer_output]
            else:
                print('Model Layer Name List: ', layer_name_list)
                print('Output layer: ', layer_output)
                raise IndexError('Layer index out of range')
        else:
            print('Output layer: ', layer_output)
            print('typeof (output layer) : ', type(layer_output))
            raise TypeError(
                '\"get_other_layer\" requires parameter "layer_output" to be int or str.')
        _dict = self.get_all_layer(x, layer_input=layer_input)
        if layer_name not in _dict.keys():
            print(_dict.keys())
        return _dict[layer_name]

    def get_layer_name(self, extra=True) -> list[str]:
        layer_name = []
        for name, _ in self.features.named_children():
            if 'relu' not in name and 'bn' not in name and 'dropout' not in name:
                layer_name.append('features.' + name)
        if extra:
            layer_name.append('pool')
            layer_name.append('flatten')
        for name, _ in self.classifier.named_children():
            if 'relu' not in name and 'bn' not in name and 'dropout' not in name:
                layer_name.append('classifier.' + name)
        return layer_name


class ImageModel(Model):

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--sgm', dest='sgm', action='store_true',
                           help='whether to use sgm gradient, defaults to False')
        group.add_argument('--sgm_gamma', dest='sgm_gamma', type=float,
                           help='sgm gamma, defaults to 1.0')
        return group

    def __init__(self, name: str = 'imagemodel', layer: int = None, default_layer: int = None,
                 model_class: type[_ImageModel] = _ImageModel, dataset: ImageSet = None,
                 sgm: bool = False, sgm_gamma: float = 1.0, **kwargs):
        name, layer = ImageModel.split_name(name, layer=layer, default_layer=default_layer)
        if layer:
            name: str = name + str(layer)
        self.layer = layer
        if 'norm_par' not in kwargs.keys() and isinstance(dataset, ImageSet):
            kwargs['norm_par'] = dataset.norm_par
        super().__init__(name=name, model_class=model_class, layer=layer, dataset=dataset, **kwargs)
        if self.num_classes is None:
            self.num_classes = 1000
        self.sgm: bool = sgm
        self.sgm_gamma: float = sgm_gamma
        self.param_list['imagemodel'] = ['sgm']
        if sgm:
            self.param_list['imagemodel'].extend(['sgm_gamma'])
        self._model: _ImageModel = self._model
        self.dataset: ImageSet = self.dataset

    def get_layer(self, x: torch.Tensor, layer_output: str = 'logits', layer_input: str = 'input') -> torch.Tensor:
        return self._model.get_layer(x, layer_output=layer_output, layer_input=layer_input)

    def get_layer_name(self, extra=True) -> list[str]:
        return self._model.get_layer_name(extra=extra)

    def get_all_layer(self, x: torch.Tensor, layer_input: str = 'input') -> dict[str, torch.Tensor]:
        return self._model.get_all_layer(x, layer_input=layer_input)

    def grad_cam(self, _input: torch.FloatTensor, _class: list[int]) -> np.ndarray:
        if isinstance(_class, int):
            _class = [_class] * len(_input)
        _class = torch.tensor(_class).to(_input.device)
        feats = self._model.get_fm(_input).detach()   # (N,C,H,W)
        feats.requires_grad_()
        _output: torch.FloatTensor = self._model.pool(feats)
        _output: torch.FloatTensor = self._model.flatten(_output)
        _output: torch.FloatTensor = self._model.classifier(_output)
        _output: torch.FloatTensor = _output.gather(dim=1, index=_class.unsqueeze(1)).sum()
        grad: torch.FloatTensor = torch.autograd.grad(_output, feats)[0]   # (N,C,H,W)
        feats.requires_grad_(False)

        weights: torch.FloatTensor = grad.mean(dim=-2, keepdim=True).mean(dim=-1, keepdim=True)    # (N,C,1,1)
        heatmap: torch.FloatTensor = (feats * weights).sum(dim=1).clamp(0)  # (N,H,W)
        heatmap.sub_(heatmap.min(dim=-2, keepdim=True)[0].min(dim=-1, keepdim=True)[0])
        heatmap.div_(heatmap.max(dim=-2, keepdim=True)[0].max(dim=-1, keepdim=True)[0])
        heatmap = (to_numpy(heatmap).transpose(1, 2, 0) * 255).astype(np.uint8)
        heatmap = Image.fromarray(heatmap).resize(_input.shape[-2:], resample=Image.BICUBIC)
        heatmap = np.array(heatmap)
        if len(heatmap.shape) == 2:
            heatmap = heatmap.reshape(heatmap.shape[0], heatmap.shape[1], 1)
        heatmap = heatmap.transpose(2, 0, 1).astype(float) / 255    # (N, H, W)
        return heatmap

    def get_saliency_map(self, _input: torch.FloatTensor, _class: list[int]) -> torch.Tensor:
        if isinstance(_class, int):
            _class = [_class] * len(_input)
        _class: torch.Tensor = torch.tensor(_class).to(_input.device)
        x: torch.FloatTensor = _input.detach()
        x.requires_grad_()
        _output: torch.FloatTensor = self(x)
        _output: torch.FloatTensor = _output.gather(dim=1, index=_class.unsqueeze(1)).sum()
        grad: torch.FloatTensor = torch.autograd.grad(_output, x)[0]   # (N,C,H,W)
        x.requires_grad_(False)

        heatmap = grad.clamp(min=0).max(dim=1)[0]   # (N,H,W)
        heatmap.sub_(heatmap.min(dim=-2, keepdim=True)[0].min(dim=-1, keepdim=True)[0])
        heatmap.div_(heatmap.max(dim=-2, keepdim=True)[0].max(dim=-1, keepdim=True)[0])
        return heatmap

    @staticmethod
    def split_name(name, layer=None, default_layer=0, output=False):
        return func(name, layer=layer, default_layer=default_layer, output=output)
